#!/usr/bin/env python

import cgi, os, sys

sys.path.append("/var/lib/imip-agent")

from imiptools.content import format_datetime, get_datetime, get_item, \
                              get_utc_datetime, get_values, parse_object, \
                              to_timezone
from imiptools.period import have_conflict
import markup
import imip_store

getenv = os.environ.get
setenv = os.environ.__setitem__

class CGIEnvironment:

    "A CGI-compatible environment."

    def __init__(self):
        self.args = None
        self.method = None
        self.path = None
        self.path_info = None
        self.user = None

    def get_args(self):
        if self.args is None:
            if self.get_method() != "POST":
                setenv("QUERY_STRING", "")
            self.args = cgi.parse(keep_blank_values=True)
        return self.args

    def get_method(self):
        if self.method is None:
            self.method = getenv("REQUEST_METHOD") or "GET"
        return self.method

    def get_path(self):
        if self.path is None:
            self.path = getenv("SCRIPT_NAME") or ""
        return self.path

    def get_path_info(self):
        if self.path_info is None:
            self.path_info = getenv("PATH_INFO") or ""
        return self.path_info

    def get_user(self):
        if self.user is None:
            self.user = getenv("REMOTE_USER") or ""
        return self.user

    def get_output(self):
        return sys.stdout

    def get_url(self):
        path = self.get_path()
        path_info = self.get_path_info()
        return "%s%s" % (path.rstrip("/"), path_info)

class Manager:

    "A simple manager application."

    def __init__(self):
        self.store = imip_store.FileStore()
        self.env = CGIEnvironment()
        user = self.env.get_user()
        self.user = user and "mailto:%s" % user or None
        self.out = self.env.get_output()
        self.page = markup.page()
        self.encoding = "utf-8"

    def new_page(self, title):
        self.page.init(title=title, charset=self.encoding)

    def status(self, code, message):
        print >>self.out, "Status:", code, message

    def no_user(self):
        self.status(403, "Forbidden")
        self.new_page(title="Forbidden")
        self.page.p("You are not logged in and thus cannot access scheduling requests.")

    def no_page(self):
        self.status(404, "Not Found")
        self.new_page(title="Not Found")
        self.page.p("No page is provided at the given address.")

    def show_requests(self):

        "Show requests for the current user."

        # NOTE: This list could be more informative, but it is envisaged that
        # NOTE: the requests would be visited directly anyway.

        self.new_page(title="Pending Requests")
        self.page.ul()

        requests = self.store.get_requests(self.user)

        for request in requests:
            self.page.li()
            self.page.a(request, href="%s/%s" % (self.env.get_url(), request))
            self.page.li.close()

        self.page.ul.close()

    def show_request(self, path_info):

        "Show a request using the given 'path_info' for the current user."

        uid = path_info.lstrip("/").split("/", 1)[0]
        f = uid and self.store.get_event(self.user, uid) or None

        if not f:
            return False

        request = parse_object(f, "utf-8")

        if not request:
            return False

        # Handle a submitted form.

        args = self.env.get_args()
        show_form = False

        if args.has_key("accept"):
            pass
        elif args.has_key("decline"):
            pass
        elif args.has_key("ignore"):
            pass
        else:
            show_form = True

        self.new_page(title="Request")

        # Provide a summary of the request.

        self.page.p("The following request was received:")
        self.page.dl()

        for name in ["SUMMARY", "DTSTART", "DTEND", "ORGANIZER", "ATTENDEE"]:
            for value in get_values(request, name):
                self.page.dt(name)
                self.page.dd(value)

        self.page.dl.close()

        dtstart = format_datetime(get_utc_datetime(request, "DTSTART"))
        dtend = format_datetime(get_utc_datetime(request, "DTEND"))

        # Indicate whether there are conflicting events.

        freebusy = self.store.get_freebusy(self.user)
        if freebusy:

            # Obtain any time zone details from the suggested event.

            _dtstart, attr = get_item(request, "DTSTART")
            tzid = attr.get("TZID")

            # Show any conflicts.

            for start, end, uid in have_conflict(freebusy, [(dtstart, dtend)], True):
                start = format_datetime(to_timezone(get_datetime(start), tzid))
                end = format_datetime(to_timezone(get_datetime(end), tzid))
                self.page.p("Event conflicts with another from %s to %s." % (start, end))

        # Show a form if no action has just been taken.

        if show_form:
            self.page.p("Action to take for this request:")
            self.page.form(method="POST")
            self.page.p()
            self.page.input(name="accept", type="submit", value="Accept")
            self.page.add(" ")
            self.page.input(name="decline", type="submit", value="Decline")
            self.page.add(" ")
            self.page.input(name="ignore", type="submit", value="Ignore")
            self.page.p.close()
            self.page.form.close()

        return True

    def select_action(self):

        "Select the desired action and show the result."

        path_info = self.env.get_path_info().rstrip("/")
        if not path_info:
            self.show_requests()
        elif self.show_request(path_info):
            pass
        else:
            self.no_page()

    def show(self):

        "Interpret a request and show an appropriate response."

        if not self.user:
            self.no_user()
        else:
            self.select_action()

        # Write the headers and actual content.

        print >>self.out, "Content-Type: text/html; charset=%s" % self.encoding
        print >>self.out
        self.out.write(unicode(self.page).encode(self.encoding))

if __name__ == "__main__":
    Manager().show()

# vim: tabstop=4 expandtab shiftwidth=4
