#!/usr/bin/env python

"""
Handlers for a resource.
"""

from imiptools.content import Handler, format_datetime, to_part
from vCalendar import to_node

class Event(Handler):

    "An event handler."

    def add(self):
        pass

    def cancel(self):
        pass

    def counter(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def declinecounter(self):

        """
        Since this handler does not send counter proposals, it will not handle
        replies to such proposals.
        """

        pass

    def publish(self):
        pass

    def refresh(self):
        pass

    def reply(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def request(self):

        """
        Respond to a request by preparing a reply containing accept/decline
        information for each indicated attendee.

        No support for countering requests is implemented.
        """

        oa = self.require_organiser_and_attendees()
        if not oa:
            return None

        (organiser, organiser_attr), attendees = oa

        # Process each attendee separately.

        calendar = []

        for attendee, attendee_attr in attendees.items():

            # Check for event using UID.

            if not self.have_new_object(attendee, "VEVENT"):
                continue

            # If newer than any old version, discard old details from the
            # free/busy record and check for suitability.

            periods = self.get_periods()
            freebusy = self.update_freebusy(attendee, periods)
            scheduled = freebusy is not None

            attendee_attr["PARTSTAT"] = scheduled and "ACCEPTED" or "DECLINED"

            self.details["ATTENDEE"] = [(attendee, attendee_attr)]

            calendar.append(to_node(
                {"VEVENT" : [(self.details, {})]}
                ))

            if scheduled:
                self.store.set_event(attendee, self.uid, to_node(
                    {"VEVENT" : [(self.details, {})]}
                    ))
                if self.publisher:
                    self.publisher.set_freebusy(attendee, freebusy)

        return "REPLY", to_part("REPLY", calendar)

class Freebusy(Handler):

    "A free/busy handler."

    def publish(self):
        pass

    def reply(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def request(self):

        """
        Respond to a request by preparing a reply containing free/busy
        information for each indicated attendee.
        """

        oa = self.require_organiser_and_attendees()
        if not oa:
            return None

        (organiser, organiser_attr), attendees = oa

        # Construct an appropriate fragment.

        calendar = []
        cwrite = calendar.append

        # Get the details for each attendee.

        for attendee, attendee_attr in attendees.items():
            freebusy = self.store.get_freebusy(attendee)

            record = []
            rwrite = record.append

            rwrite(("ORGANIZER", organiser_attr, organiser))
            rwrite(("ATTENDEE", attendee_attr, attendee))
            rwrite(("UID", {}, self.uid))

            if freebusy:
                for start, end, uid in freebusy:
                    rwrite(("FREEBUSY", {"FBTYPE" : "BUSY"}, [start, end]))

            cwrite(("VFREEBUSY", {}, record))

        # Return the reply.

        return "REPLY", to_part("REPLY", calendar)

class Journal(Handler):

    "A journal entry handler."

    def add(self):
        pass

    def cancel(self):
        pass

    def publish(self):
        pass

class Todo(Handler):

    "A to-do item handler."

    def add(self):
        pass

    def cancel(self):
        pass

    def counter(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def declinecounter(self):

        """
        Since this handler does not send counter proposals, it will not handle
        replies to such proposals.
        """

        pass

    def publish(self):
        pass

    def refresh(self):
        pass

    def reply(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def request(self):
        pass

# Handler registry.

handlers = [
    ("VFREEBUSY",   Freebusy),
    ("VEVENT",      Event),
    ("VTODO",       Todo),
    ("VJOURNAL",    Journal),
    ]

# vim: tabstop=4 expandtab shiftwidth=4
